#include <v_lib.h>
#include "rgcache.h"
#include "genparse.h"

#define	GEN_VC	0
#define	GEN_FL	1
#define	GEN_FD	2

/*
 * Filler particle: fiber.
 */
typedef struct fp_s
{
	vec3_t	p1, p2;
	vec_t	len, diameter;
	vec_t	volume;
}	fp_t, fp_p;

/*
 * Contact point
 */
typedef struct fc_s
{
	int fn1, fn2;	/* pointers to contacting fibers */
	vec_t dist;		/* minimal distance between fibers */
	vec3_t cp;		/* contact point */
}	fc_t, *fc_p;

/*
 * System to be modeled.
 */
typedef struct tsys_s
{
	vec3_t		bbox[2];
	vec_t		volume;
	vec_t		volumetric_content;
	bstack_t	fs; /* of fp_t */
	bstack_t	cs; /* of fc_t */
}	tsys_t, *tsys_p;

typedef struct tseq_s
{
	vec3_t bbox[2];
	/*
	 * Number of iterators among generators.
	 * 0 - one experiment;
	 * 1 - non-nested loop;
	 * 2 - doubly nested loop;
	 * 3 - triply-nested loop.
	 */
	u32_t nesting_level;
	/*
	 * #0 - volumetric content;
	 * #1 - fiber length;
	 * #2 - fiber diameter.
	 */
	fgen_p gens[3];
	/*
	 * Indices in 'gens' pointing to iterators.
	 * Actual number of indices is equal to 'nesting_level'.
	 */
	u32_t iterators[3];
	/*
	 * There are 2 stages of modeling process.
	 * 1. Generation of the test systems. Here non-reentrant code
	 * (generators are shared between all tasks) is used.
	 *
	 * 2. Computing number of contacts. Computation-intensive stage,
	 * subject to later parallelization.
	 *
	 * Thus, total number of test systems depends on 'nesting_level' and
	 * parameters of iterators (fg->u.ilinear.steps + 1).
	 * The seqSysAmount() is for initial determination of the former.
	 */
}	tseq_t, *tseq_p;

vboolean V_GLOBAL
parseSequence(vhandle_t lex, vp_stack_p generators, vp_stack_p store)
{
	/* FIXME: no dynamic data, use bstack_t for storage */
	CALLER_NAME(parseSequence)
	STATIC_STRING(duplicate_definition, "duplicate definition")
	STATIC_STRING(generator_undefined, "generator undefined")
	int i;
	vec_t	raw_bbox[6];
	char	vc_name[MAX_ID_LEN];
	char	len_name[MAX_ID_LEN];
	char	diam_name[MAX_ID_LEN];
	fgen_p	gens[3];
	char	*name;
#define	GOT_VC		1
#define	GOT_FL		2
#define	GOT_FD		4
#define	GOT_BBOX	8
#define	GOT_ALL		15	/* (GOT_BBOX|GOT_VC|GOT_LEN|GOT_DIAM) */
	u32_t	got_flags = 0;
	tseq_p seq;
	if(!lexNextIs(lex, "{"))
	{
		SYN_EXPECTED(lex, "{");
		return false;
	}
	while(lexNext(lex))
	{
		if(lexLastWas(lex, "}"))
		{
			if(!(got_flags & GOT_ALL))
			{
				PARSE_ERROR(lex, "too few parameters to start simulation");
				return false;
			}
			break;
		}
		else if(lexLastWas(lex, "bounding_box"))
		{
			puts("========== bounding_box");
			if(got_flags & GOT_BBOX)
			{
				PARSE_ERROR(lex, duplicate_definition);
				return false;
			}
			/* parse 6 floats, followed by ';' */
			for(i = 0;i < 6;i++)
			{
				if(!lexNextFloat(lex, raw_bbox + i))
				{
					SYN_EXPECTED(lex, "floating-point value for bbox");
					return false;
				}
			}
			if(!lexNextIs(lex, ";"))
			{
				SYN_EXPECTED(lex, ";");
				return false;
			}
			got_flags = got_flags | GOT_BBOX;
		}
		else if(lexLastWas(lex, "volumetric_content"))
		{
			puts("========== VC");
			if(got_flags & GOT_VC)
			{
				PARSE_ERROR(lex, duplicate_definition);
				return false;
			}
			name = strSafe(lexNext(lex));
			
			if(!lexNextIs(lex, ";"))
			{
				SYN_EXPECTED(lex, ";");
				return false;
			}
			strCopy(vc_name, name, MAX_ID_LEN);
			gens[GEN_VC] = genFind(generators, vc_name);
			if(!gens[GEN_VC])
			{
				PARSE_ERROR(lex, generator_undefined);
				return false;
			}
			got_flags = got_flags | GOT_VC;
		}
		else if(lexLastWas(lex, "fiber_length"))
		{
			puts("========== FL");
			if(got_flags & GOT_FL)
			{
				PARSE_ERROR(lex, duplicate_definition);
				return false;
			}
			name = lexNext(lex);
			
			
			if(!lexNextIs(lex, ";"))
			{
				SYN_EXPECTED(lex, ";");
				return false;
			}
			strCopy(len_name, name, MAX_ID_LEN);
			gens[GEN_FL] = genFind(generators, len_name);
			if(!gens[GEN_FL])
			{
				PARSE_ERROR(lex, generator_undefined);
				return false;
			}
			got_flags = got_flags | GOT_FL;
		}
		else if(lexLastWas(lex, "fiber_diameter"))
		{
			puts("========== FD");
			if(got_flags & GOT_FD)
			{
				PARSE_ERROR(lex, duplicate_definition);
				return false;
			}
			name = lexNext(lex);
			if(!lexNextIs(lex, ";"))
			{
				SYN_EXPECTED(lex, ";");
				return false;
			}
			strCopy(diam_name, name, MAX_ID_LEN);
			gens[GEN_FD] = genFind(generators, diam_name);
			if(!gens[GEN_FD])
			{
				PARSE_ERROR(lex, generator_undefined);
				return false;
			}
			got_flags = got_flags | GOT_FD;
		}
		else
		{
			PARSE_ERROR(lex, "unknown keyword");
			return false;
		}
	}
	if(!lexLast(lex))
	{
		SYN_UNEXPECTED_EOF(lex);
		return false;
	}
	V_ASSERT(got_flags & GOT_ALL);
	/*
	 * All Ok. Can do allocations.
	 */
	seq = (tseq_p)memAllocZ(sizeof(tseq_t));
	/* mins & maxs from origin & extents */
	vec3Scale(raw_bbox + 3, raw_bbox + 3, (vec_t)0.5);
	vec3Sub(seq->bbox[0], raw_bbox, raw_bbox + 3);
	vec3Add(seq->bbox[1], raw_bbox, raw_bbox + 3);
	/* store generators */
	seq->gens[0] = gens[0];
	seq->gens[1] = gens[1];
	seq->gens[2] = gens[2];
	/* find iterators and store their positions */
	seq->nesting_level = 0;
	for(i = 0;i < 3;i++)
	{
		if(IS_ITERATOR(seq->gens[i]))
			seq->iterators[seq->nesting_level++] = i;
	}
	vpsPush(store, seq);
	return true;
}

u32_t	V_GLOBAL
seqSysAmount(tseq_p seq)
{
	switch(seq->nesting_level)
	{
		case 0:
			return 1;
		case 1:
			return GEN_NUM_ITERATIONS(seq->gens[seq->iterators[0]]);
		case 2:
			return GEN_NUM_ITERATIONS(seq->gens[seq->iterators[0]]) *
				GEN_NUM_ITERATIONS(seq->gens[seq->iterators[1]]);
		case 3:
			return GEN_NUM_ITERATIONS(seq->gens[seq->iterators[0]]) *
				GEN_NUM_ITERATIONS(seq->gens[seq->iterators[1]]) *
				GEN_NUM_ITERATIONS(seq->gens[seq->iterators[2]]);
		default:
			V_ASSERT(0);
			break;
	}
	V_ASSERT(0);
	return (u32_t)(-1);
}

void	V_GLOBAL
seqDestroy(vp_stack_p store)
{
	void_p ptr;
	while(vpsPop(store, &ptr))
		memFree(ptr);
	vpsClear(store);
}

void	V_GLOBAL
seqPrint(vp_stack_p store)
{
	u32_t i, j;
	tseq_p seq;
	printf("Parsed %u sequences.\n", vpsCount(store));
	for(i = 0;i < vpsCount(store);i++)
	{
		static char *gn[3] = {"VC, FL, FD"};
		seq = (tseq_p)vpsGet(store, i);
		printf("Seq No %u of nesting_level %u:\n"
			"bounding box [%g,%g,%g]-[%g,%g,%g]\nGenerators:\n",
			i, seq->nesting_level,
			seq->bbox[1][0],seq->bbox[1][1],seq->bbox[1][2],
			seq->bbox[2][0],seq->bbox[2][1],seq->bbox[2][2]);
		for(j = 0;j < 3;j++)
			printf("%s: '%s'.\n",gn[j], seq->gens[j]->name);
		if(seq->nesting_level)
		{
			puts("Iterator(s):");
			for(j = 0;j < seq->nesting_level;j++)
				printf("%s: '%s'.\n",gn[seq->iterators[j]], seq->gens[seq->iterators[j]]->name);
		}
		printf("Total number of experiments: %u\n", seqSysAmount(seq));
	}
}

vboolean	V_GLOBAL
sysCreateSystems(vp_stack_p seqs, vp_stack_p store)
{
	return false;
}

/*
 * TODO for near holiday: sysParallelScheduler().
 */
vboolean	V_GLOBAL
sysSequentialScheduler(vp_stack_p syss)
{
	return false;
}

/*
 * Saving the results, visualization, etc.
 */
vboolean	V_GLOBAL
sysPostprocess(vp_stack_p syss)
{
	return false;
}

void	V_GLOBAL
sysDestroy(vp_stack_p syss)
{
}

vhandle_t lex;
vp_stack_t gens = TYPED_STACK_EMPTY_INITIALIZER;
vp_stack_t seqs = TYPED_STACK_EMPTY_INITIALIZER;
vp_stack_t syss = TYPED_STACK_EMPTY_INITIALIZER;

int main(int argc, char *argv[])
{
	CALLER_NAME(main)
	char *script = (argc > 1) ? argv[1] : "test.fcic";
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	lex = lexParseFile(script, "{}()[];");
	if(!lex)
		return -1;
	lexSetInteger(lex, LEX_COMMENT_CHAR, '#');
	lexEnable(lex, LEX_ONE_CHAR_COMMENTS);
	while(lexNext(lex))
	{
		if(lexLastWas(lex, "generators"))
		{
			parseGenerators(lex, &gens);
		}
		else if(lexLastWas(lex, "sequence"))
		{
			if(!parseSequence(lex, &gens, &seqs))
			{
				puts("ERR PA SEQ");
				goto parse_error;
			}
		}
		else
			goto parse_error;
	}
	puts("PARSED");
	/*
	if(sysCreateSystems(&seqs, &syss))
		if(sysSequentialScheduler(&syss))
			sysPostprocess(&syss);
	*/
	goto cleanup;
parse_error:;
	PARSE_ERROR(lex, "generic");
cleanup:;
    genDestroy(&gens);
    seqDestroy(&seqs);
    sysDestroy(&syss);
	lexDestroy(lex);
	memPrintStats();
	return 0;
}
